#!/usr/bin/env python3
#
# Copyright (c) 2022 waverian. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import argparse
import re
import webbrowser
import logging
import sys
from typing import Optional


def setup_logger(verbosity: int = None, log_file: Optional[str] = None) -> None:
    """"Set up the logger with specified verbosity level and log file, with different formats for each handler."""
    handlers = []
    stdout_level = {
        0: logging.WARN,
        1: logging.INFO,
        2: logging.DEBUG
    }.get(min(verbosity, 2), logging.WARN)

    # Create handlers
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(stdout_level)
    stdout_formatter = logging.Formatter('%(message)s')
    stdout_handler.setFormatter(stdout_formatter)
    stdout_handler.addFilter(lambda record: record.levelno <= logging.WARNING)
    handlers.append(stdout_handler)

    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.setLevel(logging.ERROR)
    stderr_formatter = logging.Formatter('%(message)s')
    stderr_handler.setFormatter(stderr_formatter)
    handlers.append(stderr_handler)

    if log_file:
        file_handler = logging.FileHandler(log_file)
        file_handler.setLevel(logging.DEBUG)
        file_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                                           datefmt='%Y-%m-%d %H:%M:%S')
        file_handler.setFormatter(file_formatter)
        handlers.append(file_handler)

    logging.basicConfig(
        level=logging.NOTSET,
        handlers=handlers
    )


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description='Script')

    parser.add_argument('path', type=str, nargs='?', default='.',
                        help='Inputh path to the folder with benchmark results')
    parser.add_argument('--recursive', action='store_true', help='Enable recursive walk mode')
    parser.add_argument('--sort', '-s',
                        help='Specify sorting type: "st" (single-threaded) or "mt" (multi-threaded). Defaults to "st".',
                        type=str, choices=['st', 'mt'], default='st'
                        )

    logging_group = parser.add_argument_group('logging options')
    logging_group.add_argument('-v', '--verbose', help='Increase verbosity (use -v, -vv)',
                               action='count', default=0, dest='log_level')
    logging_group.add_argument('-l', '--log_file', help='Set path to output log file',
                               action='store', default=None)

    return parser.parse_args()


########################################################################################################################
########################################################################################################################
########################################################################################################################

import os


def read_file(filepath):
    encodings = ["utf-8", "cp1252", "iso-8859-1"]  # List of most common encodings
    for enc in encodings:
        try:
            with open(filepath, "r", encoding=enc, errors="replace") as file:
                return file.read()
        except UnicodeDecodeError:
            continue
    raise ValueError("File encoding could not be determined.")


def app_main(args: argparse.Namespace) -> int:
    path = os.path.abspath(os.path.expanduser(args.path))
    logging.info(f'Base path is {path}')
    if not os.path.isdir(path):
        raise RuntimeError(f'Input path {path} is not a valid directory')

    all_files = []
    for dirpath, _, filenames in os.walk(path):
        for filename in filenames:
            all_files.append(os.path.join(dirpath, filename))
        if not args.recursive:
            break

    logging.debug(f'Change current working directory to {path}')
    os.chdir(path)

    reports = dict()
    duplicated_report_names = dict()
    for file_full_path in all_files:
        file_name = os.path.basename(file_full_path)

        report_name = os.path.splitext(file_name)[0]
        extension = os.path.splitext(file_name)[1]

        if extension.lower() == '.txt':
            key = 'text'
        elif extension.lower() in ['.html', '.htm']:
            key = 'html'
        else:
            logging.info(f'File {file_full_path} has unsupported extension {extension.lower()}. Ignoring.')
            continue

        if report_name in duplicated_report_names:
            duplicated_report_names[report_name].add(file_full_path)
            continue

        if not report_name in reports:
            reports[report_name] = {}

        if key in reports[report_name]:
            duplicated_report_names[report_name] = set(reports[report_name].values())
            duplicated_report_names[report_name].add(file_full_path)
            del reports[report_name]
            continue

        reports[report_name][key] = os.path.relpath(file_full_path)

    reports_to_delete = set()
    for report_name, report in reports.items():
        for required_file in ['text', 'html']:
            if required_file not in report:
                logging.info(f'{report_name} has no {required_file} report. Ignoring.')
                reports_to_delete.add(report_name)

    for report_name in reports_to_delete:
        del reports[report_name]

    if duplicated_report_names:
        logging.warning(f'Some report names are duplicated, so will be excluded.')
        for report_name, files in duplicated_report_names.items():
            files_fmt = '\n  '.join(files)
            logging.info(f"{report_name} is duplicated:\n  {files_fmt}")

    reports_to_delete = set()
    for report, data in reports.items():
        logging.debug(f'Processing report {report}')
        report_text = read_file(data['text'])

        cpu_name_match = re.search(r'(?:cpu name).*?- (.*)', report_text, flags=re.IGNORECASE)
        data['cpu_name'] = cpu_name_match.groups()[0]
        st_result_match = re.search(r'(?:singlecore result|single core).*?- ([\d\.]*)', report_text,
                                    flags=re.IGNORECASE)
        data['st_result'] = float(st_result_match.groups()[0])
        mt_result_match = re.search(r'(?:multicore result|multi core).*?- ([\d\.]*)', report_text, flags=re.IGNORECASE)
        data['mt_result'] = float(mt_result_match.groups()[0])

    for report_name in reports_to_delete:
        del reports[report_name]

    sorted_data = dict(sorted(reports.items(), key=lambda item: item[1][f'{args.sort}_result'] if item[1][
        f'{args.sort}_result'] else float('-inf'), reverse=True))

    with open('index.html', 'w+', encoding='utf-8') as file:
        py_print_html(file, sorted_data)

    webbrowser.open('index.html')
    return 0


def py_print_html(file, reports):
    file.write(
        '<html xmlns=\"http://www.w3.org/1999/html\">\n<head>\n\t<meta charset=\"utf-8\">\n\t<title>Waverian Evolution benchmark results summary</title>\n\t<style>\n        body {\n            font-family: system-ui, -apple-system, \"Segoe UI\", Roboto, \"Helvetica Neue\", Arial, \"Noto Sans\", \"Liberation Sans\", sans-serif, \"Apple Color Emoji\", \"Segoe UI Emoji\", \"Segoe UI Symbol\", \"Noto Color Emoji\"\n        }\n\n        .container {\n            margin-left: auto;\n            margin-right: auto;\n            min-width: 30em;\n            max-width: 50em;\n            text-align: center;\n        }\n\n        .thin_font {\n            font-weight: 300;\n            margin: auto;\n        }\n\n        .white_on_blue {\n            background-color: #1e90ff;\n            color: #f0f8ff;\n        }\n\n\t\ttable.blue_border td{\n\t\t\tborder: 1px solid #1e90ff;\n\t\t}\n\n        h3 {\n            font-size: 1.8em;\n        }\n\n        table {\n            border-collapse: collapse;\n            border-width: 0;\n            width: 100%;\n            text-align: left;\n        }\n\n        p {\n            text-align: left;\n        }\n\n        .logo {\n            text-align: left;\n            padding: 10px;\n        }\n\t</style>\n    <script>\n        document.addEventListener(\"DOMContentLoaded\", () => {\n            const table = document.getElementById(\"score_table\");\n            const headers = table.querySelectorAll(\"th\");\n            const tbody = table.querySelector(\"tbody\");\n            let currentSort = { index: 0, direction: \"asc\" };\n\n            const sortTable = (index, direction) => {\n                const rows = Array.from(tbody.querySelectorAll(\"tr\"));\n                rows.sort((a, b) => {\n                    const aText = a.cells[index].textContent;\n                    const bText = b.cells[index].textContent;\n\n                    return aText.localeCompare(bText, undefined, {numeric: true, sensitivity: \'base\'}) * ( direction === \"asc\" ? -1 : 1)\n                });\n\n                tbody.innerHTML = \"\";\n                rows.forEach(row => tbody.appendChild(row));\n            };\n\n            const updateHeaders = (index, direction) => {\n                headers.forEach((header, i) => {\n                    const indicator = header.querySelector(\".sort-indicator\");\n                    indicator.textContent = \"◃\";\n                    if (i === index) {\n                        indicator.textContent = direction === \"asc\" ? \"▲\" : \"▼\";\n                    }\n                });\n            };\n\n            headers.forEach((header, index) => {\n                if (!header.querySelector(\".sort-indicator\")){\n                    return\n                }\n\n                header.addEventListener(\"click\", () => {\n                    const direction = currentSort.index === index && currentSort.direction === \"asc\" ? \"desc\" : \"asc\";\n                    currentSort = { index, direction };\n                    sortTable(index, direction);\n                    updateHeaders(index, direction);\n                });\n            });\n\n            // Default sort on load\n            const defaultSortColumn = 0; // Change to desired default column index\n            const defaultSortType = headers[defaultSortColumn].dataset.type;\n            sortTable(defaultSortColumn, defaultSortType, \"asc\");\n            updateHeaders(defaultSortColumn, \"asc\");\n        });\n    </script>\n</head>\n<body>\n<div class=\"container\">\n\t<div class=\"logo white_on_blue\">\n\t\t<h3 class=\"thin_font\">Waverian Evolution benchmark results</h3>\n\t</div>\n    <br />\n\t<table class=\"blue_border\" id=\"score_table\">\n\t\t<thead class=\"white_on_blue\">\n            <tr>\n                <th data-type=\"string\">Name<span class=\"sort-indicator\">◃</span></th>\n                <th data-type=\"number\">Score ST<span class=\"sort-indicator\">◃</span></th>\n                <th data-type=\"number\">Score MT<span class=\"sort-indicator\">◃</span></th>\n                <th>Text</th>\n                <th>Html</th>\n            </tr>\n        </thead>\n\t\t<tbody>\n')

    for name, report in reports.items():
        file.write('\n            <tr>\n                <td>')
        file.write(f'{report["cpu_name"]:}')
        file.write('</td>\n                <td>')
        file.write(
            f'{report["st_result"]:.{(1 > report["st_result"]) + (10 > report["st_result"]) + (100 > report["st_result"])}f}')
        file.write('</td>\n                <td>')
        file.write(
            f'{report["mt_result"]:.{(1 > report["mt_result"]) + (10 > report["mt_result"]) + (100 > report["mt_result"])}f}')
        file.write('</td>\n                <td><a href=\"')
        file.write(f'{report["text"]:}')
        file.write('\">Text</a></td>\n                <td><a href=\"')
        file.write(f'{report["html"]:}')
        file.write('\">Html</a></td>\n            </tr>\n')
    file.write('\n\t\t</tbody>\n\t</table>\n</div>\n</body>\n</html>')


########################################################################################################################
########################################################################################################################
########################################################################################################################

def main():
    __arguments = parse_arguments()
    setup_logger(__arguments.log_level, __arguments.log_file)
    del __arguments.log_level, __arguments.log_file
    __error_code = 0
    try:
        logging.debug(f'ENTER main with command line argumetns {__arguments.__dict__}')
        __result = app_main(__arguments)
        __error_code = __result if __result is int else 0
    except Exception as e:
        logging.fatal(f'Fatal error occurred:', exc_info=True)
        __error_code = 1
        raise e
    finally:
        logging.debug(f'EXIT main with error code {__error_code}')
        exit(__error_code)


if __name__ == '__main__':
    main()
